import bpy
from bpy.types import Operator
from bpy.props import *
from mathutils import Matrix

from .utils import (
    icon
)
from .trail_generator import setup_trail, setup_trail_armature

def draw_operator( layout, cls, text = None, icon = 'NONE', icon_value = 0, overrides = {}):

    row = layout.row( align = True )
    for i in list( overrides ):
        row.context_pointer_set( i, overrides[ i ])
    op = row.operator( cls.bl_idname, text = text if not text == None else cls.bl_label, icon = icon, icon_value = icon_value )
    return op

class MAGICTRAILS_OT_new_trail( Operator ):

    bl_idname = "magictrails.new_trail"
    bl_label = "Add New Trail"
    bl_options = {'UNDO'}
    bl_description = 'Add a new trail to your scene.'

    root : EnumProperty( items= [
        ( 'CURSOR', '3D Cursor', 'Spawn Trail at 3D cursor' ),
        ( 'OBJECT', 'Object', 'Attach new Trail to active object' ),
        ( 'BONE', 'Bone', 'Attach new Trail to active bone' )
    ],
    default = 0,
    name = 'Root'
    )
    use_local_bone_coords : BoolProperty( default = False, name = 'Test' )

    @classmethod
    def poll( cls, context ):
        return True

    def draw( self, context ):

        settings = context.scene.mtrail_settings
        layout = self.layout
        box = layout.box( )
        settings.draw( box )

    def execute( self, context ):

        obj = context.object
        settings = context.scene.mtrail_settings
        controller = bpy.data.objects.new( settings.name, None )
        if self.root == 'CURSOR':
            mat = context.scene.cursor.matrix
            controller.matrix_world = mat
        elif self.root == 'OBJECT' and obj:
            controller.parent = obj
        elif self.root == 'BONE' and obj.type == 'ARMATURE' and obj.data.bones.active:
            bone = obj.data.bones.active
            controller.parent = obj
            controller.parent_type = 'BONE'
            controller.parent_bone = bone.name
            if self.use_local_bone_coords:
                mat = Matrix((
                    ( 1, 0, 0, 0 ),
                    ( 0, 0, 1, -1 ),
                    ( 0, -1, 0, 0 ),
                    ( 0, 0, 0, 1 )
                ))
                controller.matrix_basis = mat
        else:
            print( f'Could not find object associated with root type:"{self.root}"')
            return{'CANCELLED'}

        setup_trail( controller, settings )

        return {'FINISHED'}

    def invoke( self, context, event ):

        return context.window_manager.invoke_props_dialog( self )

class MAGICTRAILS_OT_new_trail_on_controller( Operator ):
    bl_idname = "magictrails.new_trail_on_controller"
    bl_label = "Add Trail Mesh"
    bl_options = {'UNDO'}
    bl_description = 'Attach a new Trail mesh to your controller'

    @classmethod
    def poll( cls, context ):
        return context.object and context.object.mtrail.is_controller

    def execute( self, context ):

        c = context.object
        c_settings = c.mtrail.settings
        defaults = context.scene.mtrail_settings
        defaults.name = c.name
        defaults.trail_resolution = c_settings.cache_length
        defaults.generate_mesh = True
        setup_trail( c, defaults )
        return{'FINISHED'}

class MAGICTRAILS_OT_set_material( Operator ):

    bl_idname = "magictrails.set_material"
    bl_label = "Change Material"
    bl_options = {'UNDO'}
    bl_description = 'Set the material of the current trail mesh. Uses presets and available materials'

    @classmethod
    def poll( cls, context ):
        return True #IMPLIES THAT THE OPERATOR ONLY GETS CALLED BY A VALID OBJECT

    def invoke( self, context, event ):
        self._obj = context.object
        scene = context.scene
        scene.mtrail_settings.set( 'material', self._obj.active_material )
        return context.window_manager.invoke_props_dialog( self )
    
    def draw( self, context ):

        layout = self.layout
        settings = context.scene.mtrail_settings
        settings.draw_material( layout ) 
        
    def execute( self, context ):
        
        settings = context.scene.mtrail_settings
        self._obj.mtrail.settings.update_material( settings.get_material( ))
        return {"FINISHED"}

class MAGICTRAILS_OT_python_allow( Operator ):

    bl_idname = "magictrails.python_allow"
    bl_label = "Reload"
    bl_options = {'UNDO'}
    bl_description = 'Allow Python to auto-execute scripts'

    def execute( self, context ):

        context.preferences.filepaths.use_scripts_auto_execute = True
        return {"FINISHED"}

class MAGICTRAILS_OT_bake( Operator ):
    bl_idname = 'magictrails.bake'
    bl_label = 'Bake Active Trail'
    bl_options = {'UNDO'}
    bl_description = 'Bake or clear trail motion'

    clear_bake : BoolProperty( name = 'Clear Bake', default = False )

    @classmethod
    def poll( cls, context ):
        return True
    
    def execute(self, context ):
        
        settings = context.object.mtrail.settings
        scene = context.scene
        if not self.clear_bake:
            settings.bake_in_range( scene.frame_start, scene.frame_end )
        else:
            settings.clear_bake( )
        return{"FINISHED"}

class MAGICTRAILS_OT_bake_to_armature( Operator ):
    bl_idname = 'magictrails.bake_to_armature'
    bl_label = 'Bake To Armature'
    bl_options = {'UNDO'}
    bl_description = 'Generate Controller Armature for Trail mesh'

    clear_bake : BoolProperty( name = 'Clear Bake', default = False )

    @classmethod
    def poll( cls, context ):
        return context.object.mtrail.type == 'MESH'
    
    def execute(self, context ):
        
        settings = context.object.mtrail.settings
        scene = context.scene
        if not self.clear_bake:
            settings.bake_in_range( scene.frame_start, scene.frame_end )
        else:
            settings.clear_bake( )

        return {'FINISHED'}

class MAGICTRAILS_OT_delete( Operator ):
    bl_idname = 'magictrails.delete'
    bl_label = 'Delete Trail'
    bl_options = {'UNDO'}
    bl_description = 'Cleanly delete active Magic Trail. Hold shift to delete single component.'

    delete_hierarchy : BoolProperty( default = True )

    @classmethod
    def poll( cls, context ):
        return context.object and context.object.mtrail.enabled
    
    def execute(self, context ):
        
        context.object.mtrail.settings.free( delete_children = self.delete_hierarchy )
        return {'FINISHED'}
    
    def invoke( self, context, event ):

        self.delete_hierarchy = not event.shift
        return self.execute( context )

class MAGICTRAILS_OT_follow_controller( Operator ):
    bl_idname = 'magictrails.follow_controller'
    bl_label = '绑定到控制器'
    bl_description = 'Make selected follow active Magic Trails Controller'

    @classmethod
    def poll( cls, context ):
        return len( context.selected_objects ) > 1 and context.object.mtrail.is_controller

    def execute(self, context ):

        for o in ( x for x in context.selected_objects if not x == context.object ):
            o.mtrail.convert_to( 'FOLLOWER', controller = context.object )
        return{'FINISHED'}

class MAGICTRAILS_OT_convert( Operator ):
    bl_idname = 'magictrails.convert'
    bl_label = 'Convert to Magic Trails'
    bl_description = 'Set up the object to be used as a Magic Trails object (Controllers, Followers)'

    convert_type : StringProperty( name = 'Convert Type', default = 'CONTROLLER' )

    @classmethod
    def poll( cls, context ):
        return context.object
    
    def execute( self, context ):

        obj = context.object
        obj.mtrail.convert_to( self.convert_type )
        return{'FINISHED'}

def add_trail_menu_item( self, context ):

    layout = self.layout
    layout.separator( )
    row = layout.row( )
    row.operator_context = 'INVOKE_DEFAULT'
    op = row.operator( MAGICTRAILS_OT_new_trail.bl_idname, text = 'New Magic Trail', icon_value = icon( 'TRAIL' ))

def follow_controller_menu_item( self, context ):

    layout = self.layout
    layout.separator( )
    row = layout.row( )
    op = row.operator( MAGICTRAILS_OT_follow_controller.bl_idname, icon_value = icon( 'cTRAIL' ))

def get_classes( ):

    yield MAGICTRAILS_OT_new_trail
    yield MAGICTRAILS_OT_bake
    yield MAGICTRAILS_OT_set_material
    yield MAGICTRAILS_OT_python_allow
    yield MAGICTRAILS_OT_delete
    yield MAGICTRAILS_OT_bake_to_armature
    yield MAGICTRAILS_OT_follow_controller
    yield MAGICTRAILS_OT_convert
    yield MAGICTRAILS_OT_new_trail_on_controller