import bpy
from bpy.types import PropertyGroup
from bpy.props import *
from mathutils import Vector, Matrix

from time import perf_counter

from .settingutils import cache_attr
from .trailsettings import MagicTrailsSettings

from ..utils import icon
from ..operators import *
from ..operators import draw_operator as op

def t( ):
    return perf_counter( ) * 10 ** 6

class ControllerSettings( MagicTrailsSettings, PropertyGroup ):

    cache_length : IntProperty( default = 3, min = 0 ) 
    is_baked : BoolProperty( default = False )

    @property
    def controller( self ):
        return self.id_data
    
    @property
    def active( self ):
        return self.cache_length > 1
    
    @property
    def children( self ):
        return [ x for x in bpy.data.objects if x.mtrail.type != 'CONTROLLER' and getattr( x.mtrail.settings, 'controller', None) == self.id_data ]

    @property
    def mesh_children( self ):
        return [ x for x in bpy.data.objects if x.mtrail.type == 'MESH' and getattr( x.mtrail.settings, 'controller', None ) == self.id_data ]
    
    def get_frame_speed( self, smoothness = 1, loc_offset = Vector(( 0, 0, 0.5 ))):

        mat_current = self.get_data( 0 )
        mat_compare = self.get_data( min( 1, smoothness ))
        loc_current = mat_current @ loc_offset
        loc_compare = mat_compare @ loc_offset
        speed = ( loc_current - loc_compare ).length / smoothness
        return speed
    
    def get_data( self, frame_offset ):

        obj = self.id_data
        cache = self.get_cache( )
        mat = Matrix( cache[ max( 0, min( frame_offset, len( cache ) - 1 )) ] )
        return mat
    
    def get_cache( self ):

        obj = self.id_data
        return obj.get( cache_attr, [ obj.matrix_world ])
    
    def realtime_update( self ):
        
        obj = self.id_data
        cache = []
        scene = bpy.context.scene
        if not scene.frame_start == scene.frame_current:
            cache = obj.get( cache_attr, [ ])
        
        mat = Matrix( obj.matrix_world )
        cache.insert( 0, mat )
        obj[ cache_attr ] = cache[ :self.cache_length ] 
        #obj.speed = self.get_frame_speed( )
        return obj[ cache_attr ]

    def bake_in_range( self, start, end ):
        
        obj = self.id_data
        override = {}
        override['object'] = obj
        override['selected_objects'] = [obj]
        override['editable_selected_objects'] = [obj]

        bpy.ops.nla.bake( override,
            frame_start = start, 
            frame_end = end,
            visual_keying = True,
            clear_constraints = True, 
            clear_parents = True,
            bake_types = { 'OBJECT' }
            )
        
        self.is_baked = True

    def draw_layout( self, layout, show_options = False, draw_children = False, box = True ):

        obj = self.id_data
        if box:
            layout = layout.box( )
        c = layout.column( align = False )
        self.draw_info( c, icon_value = icon( 'TRAIL_CONTROLLER' ))
        if not len( self.children ):
            c.prop( self, 'cache_length', text = 'Length' )
        if not len( self.mesh_children ):
            op( c, MAGICTRAILS_OT_new_trail_on_controller, icon_value = icon( 'TRAIL_MESH' ))
        if show_options:
            c_options = c.column( align = True )
            self.draw_bake( c_options )
        
        if draw_children:
            children = [ x for x in self.children  ]
            if len( children ) > 1:
                c.label( text = 'Linked:' )
            for child in children:
                child.mtrail.settings.draw_layout( c, show_options = show_options, box = box )

def get_classes( ):
    yield ControllerSettings