import bpy
from bpy.types import PropertyGroup, Object
from bpy.props import *
from mathutils import Vector

from .trailsettings import MagicTrailsSettings
from .settingutils import update_is_baked, set_controller, tranlsations, track_const_name

from ..utils import icon

def get_follow_mode_items( self, context ):

    items = [
        ( 'OFFSET', 'Offset', 'Use historical position of controller' ),
        ( 'FOLLOW_THROUGH', 'Follow-Through', 'Use historical position of controller for follow-through motion' ),
        ( 'WIGGLE', 'Wiggle', 'Use position of controller to accelerate towards' ),
        ( 'VELOCITY', 'Velocity', 'Display speed and direction in units per frame' ),
        ( 'VELOCITY_TIME', 'Real-time velocity', 'Display speed and direction in units per second' ),
    ]
    return items

class FollowerSettings( MagicTrailsSettings, PropertyGroup ):

    def update_prop( self, context ):
        self.setup_constraint( )
        self.realtime_update( )

    controller : PointerProperty( type = Object )
    is_baked : BoolProperty( name = 'Is Baked', default = False, update = update_is_baked )
    mode : EnumProperty( items = get_follow_mode_items, name = 'Mode', update = update_prop )

    @property
    def controller( self ):
        return self.controller

    @controller.setter
    def controller( self, value ):
        set_controller( self, value )

    offset : IntProperty( name = 'Frame Offset', soft_min = 1, default = 1, update = update_prop )

    #OFFSET MODE PROPERTIES
    use_loc : BoolProperty( default = True, name = 'Loc:', update = update_prop )
    use_rot : BoolProperty( default = True, name = 'Rot:', update = update_prop )
    use_scale : BoolProperty( default = True, name = 'Scale:', update = update_prop )

    #FOLLOWTHROUGH MODE PROPERTIES
    velocity : FloatVectorProperty( size = 3, name = 'Velocity', default = ( 0, 0, 0 )) #PRIVATE
    delay : IntProperty( name = 'Delay', min = 0, default = 10 )
    delay_counter : IntProperty( default = 0 ) #PRIVATE

    #WIGGLE MODE PROPERTIES
    wiggle : FloatProperty( name = 'Wiggle', subtype = 'FACTOR', min = 0, max = 1, default = 0.7 )

    def reset( self ):
        obj = self.id_data
        self.velocity = ( 0, 0, 0 )
        self.delay_counter = 0
        obj.location = self.controller.mtrail.settings.get_data( 0 ).translation

    def realtime_update( self ):
        
        if self.is_baked:
            return
        c_settings = self.controller.mtrail.settings
        mat_offset = c_settings.get_data( self.offset )
        mat_current = c_settings.get_data( 0 )
        if self.mode == 'OFFSET':
            self.set_transforms( mat_offset )
        elif self.mode == 'FOLLOW_THROUGH':
            self.set_transforms_follow_through( mat_current, mat_offset )
        elif 'VELOCITY' in self.mode:
            self.set_transforms_velocity( mat_offset, mat_current )
        else:
            self.set_transforms_wiggle( mat_current )
    
    def set_transforms_velocity( self, mat, mat2 ):

        obj = self.id_data
        loc1 = mat.translation
        obj.location = loc1
        diff = loc1 - mat2.translation
        speed = diff.length / self.offset
        if self.mode == 'VELOCITY_TIME':
            speed = speed * bpy.context.scene.render.fps
        obj.scale = [ speed ] * 3

    def set_transforms_follow_through( self, tarmat, offmat ):
        
        lerp = 0.2
        vel = Vector( self.velocity )
        obj = self.id_data

        if self.range_start( ):
            self.reset( )
            return

        obmat = obj.matrix_world
        obloc = obmat.translation
        offloc = offmat.translation
        tarloc = tarmat.translation

        if offloc == tarloc:
            lerp = min( 1, ( self.delay_counter / self.delay ))
            tarvel = vel
            obj.location = tarloc * lerp + ( obloc + tarvel ) * ( 1 - lerp )
            self.delay_counter += 1
        else:
            tarvel = ( offloc - obloc )
            self.delay_counter = 0
            obj.location = obloc + tarvel
        
        self.velocity = tarvel
    
    def set_transforms_wiggle( self, tarmat ):

        vel = Vector( self.velocity )
        obj = self.id_data

        if self.range_start( ):
            self.reset( )
            return
        
        obmat = obj.matrix_world
        obloc = obmat.translation
        tarloc = tarmat.translation
        tarvel = ( tarloc - obloc )
        lerp = self.wiggle
        tarvel = ( tarvel ) * ( 1 - lerp ) + vel * lerp
        obj.location = obloc + tarvel
        self.velocity = tarvel

    def bake_in_range( self, start, end ):
        
        obj = self.id_data
        scene = bpy.context.scene
        mem_frame = scene.frame_current
        for f in range( start, end + 1 ):
            scene.frame_set( f )
            for t in tranlsations:
                obj.keyframe_insert( t, frame = scene.frame_current )
        
        self.is_baked = True
        scene.frame_current = mem_frame

    def draw_layout( self, layout, show_options = False, draw_children = False, box = True ):

        obj = self.id_data
        if box:
            layout = layout.box( )
        c = layout.column( align = False )
        self.draw_info( c, icon_value = icon( 'TRAIL_FOLLOWER' ))
        if not self.is_baked:
            c.prop( self, 'mode' )
            if not self.mode == 'WIGGLE':
                c.prop( self, 'offset' )
            if show_options:
                if self.mode == 'OFFSET':
                    row = c.row( align = True )
                    for attr in ( 'use_loc', 'use_rot', 'use_scale' ):
                        row.prop( self, attr, toggle = 1 )
                elif self.mode == 'FOLLOW_THROUGH':
                    row = c.row( align = True )
                    row.prop( self, 'delay' )
                elif self.mode == 'WIGGLE':
                    row = c.row( align = True )
                    row.prop( self, 'wiggle' )

        self.draw_bake( c )
    
    def setup_constraint( self ):

        obj = self.id_data
        constraints = obj.constraints
        has_const = track_const_name in constraints.keys( )
        if not 'VELOCITY' in self.mode:
            if has_const:
                constraints.remove( constraints[ track_const_name ])
        else:
            if not has_const:
                c = constraints.new( 'DAMPED_TRACK' )
                c.name = track_const_name
                c.target = self.controller
                c.track_axis = 'TRACK_Z'

def get_classes( ):
    yield FollowerSettings