import bpy
from bpy.types import PropertyGroup, Material, Object, Scene
from bpy.props import *

from .controller import ControllerSettings
from .trailmesh import TrailMeshSettings
from .follower import FollowerSettings

from ..utils import get_external_materials, icon, materials_file
from ..operators import *
from ..operators import draw_operator as op

def preset_material_names( ):

    dns = bpy.app.driver_namespace
    name = 'magic_trails_preset_materials'
    if not name in dns:
        mats = list( get_external_materials( ))
        dns[name] = list( mats )
    else:
        mats = dns[name]
    for m in mats:
        yield m

def get_preset_materials( self, context ):

    items = []
    for i, m in enumerate( preset_material_names( )):
        items.append(( m, m, f'Use Preset Material "{m}"', icon( m.replace( ' ', '' ).upper( )), i ))
    return items

class TrailDefaults( PropertyGroup ):

    name : StringProperty( name = "Name", default = "MagicTrail" )
    trail_resolution : IntProperty( name = 'Length', default = 10, min = 3, soft_max = 100 )

    generate_mesh : BoolProperty( name = 'Generate Mesh', default = True, description = 'Generate a Trail Mesh when creating a Trail' )

    use_preset_material : BoolProperty( name = 'Use Preset Material', default = True )
    material : PointerProperty( name = 'Material', type = Material )
    preset_material : EnumProperty( name = 'Preset Material', items = get_preset_materials, default = 3 )

    use_local_bone_coords : BoolProperty( default = True, name = 'Local Coords' )

    def set( self, attr, value ):

        setattr( self, attr, value )

    def draw( self, l ):
        
        row = l.split( align = True )
        c = row.column( align = False )
        c.label( text = 'Name:' )
        c.prop( self, 'name', text = '' )
        split = c.split( align = True, factor = 0.85 )
        split.prop( self, 'trail_resolution' )
        split.prop( self, 'generate_mesh', text = '', toggle = 1, icon_value = icon( 'TRAIL_TIME' ))
        self.draw_material( row, enabled = self.generate_mesh )
    
    def draw_material( self, l, enabled = True ):
        
        layout = l.column( align = True )
        layout.enabled = enabled
        layout.prop( self, 'use_preset_material' )
        if self.use_preset_material:
            c = layout.column( align = True )
            c.template_icon_view( self, 'preset_material', show_labels = True, scale = 5 )
            b = c.box( )
            b.scale_y = 0.5
            b.label( text = self.preset_material )
        else:
            layout.template_ID_preview( self, 'material' )

    def get_material( self ):

        mat_data = bpy.data.materials
        if not self.use_preset_material:
            return self.material
        else:
            mat_name = self.preset_material
            if not mat_name in mat_data.keys( ):
                with bpy.data.libraries.load( materials_file, link = False ) as ( data_from, data_to ):
                    if mat_name in data_from.materials:
                        data_to.materials.append( mat_name )
            return mat_data[ mat_name ]

def get_magic_trails_types( self, context ):

    return[
        ( 'NONE', 'None', 'None' ),
        ( 'CONTROLLER', 'Controller', 'Controller' ),
        ( 'MESH', 'Mesh', 'mesh' ),
        ( 'FOLLOWER', 'Follower', 'Follower' )
    ]

class MagicTrailsProperty( PropertyGroup ):

    type : EnumProperty( items = get_magic_trails_types, name = 'Type', default = 0 )
    controller_settings : PointerProperty( type = ControllerSettings )
    mesh_settings : PointerProperty( type = TrailMeshSettings )
    follower_settings : PointerProperty( type = FollowerSettings )

    @property
    def enabled( self ):
        return self.type != 'NONE'

    @property
    def is_controller( self ):
        return self.type == 'CONTROLLER'

    @property
    def settings( self ):
        
        if self.type == 'NONE':
            return None
        else:
            d = { 'CONTROLLER':'controller_settings', 'MESH':'mesh_settings', 'FOLLOWER':'follower_settings' }
            return getattr( self, d[self.type] )
    
    @property
    def speed( self ):

        if self.type == 'NONE':
            return ( 0, 0, 1 )
        else:
            c = self.settings.controller
            r = c.frame_speed
            g = c.real_speed
            b = 0
            return ( r, g, b )

    def convert_to( self, new_type, controller = None ):

        self.type = new_type
        if new_type in ( 'MESH', 'FOLLOWER' ):
            
            if not controller:
                print( f'Could not convert to "{new_type}". Controller undefined' )
                self.type = 'NONE'
                return
            else:
                self.settings.controller = controller
            self.settings.realtime_update( )
    
    def draw_layout_default( self, layout ):

        if self.enabled:
            self.settings.draw_layout( layout, show_options = True, draw_children = True, box = True )
        c = layout.column( )

        op( c, MAGICTRAILS_OT_new_trail, text = 'New Trail on Object', icon_value = icon( 'TRAIL_TIME' )).root = 'OBJECT'
        if self.id_data.type == 'ARMATURE' and self.id_data.data.bones.active:
            l_bone = c.split( align = True, factor = 0.7 )
            bone_op = op( l_bone, MAGICTRAILS_OT_new_trail, text = 'New Trail on Bone', icon_value = icon( 'TRAIL_TIME' ))
            bone_op.root = 'BONE'
            defaults = bpy.context.scene.mtrail_settings
            bone_op.use_local_bone_coords = defaults.use_local_bone_coords
            l_bone.prop( defaults, 'use_local_bone_coords', toggle = 1 )

        if self.is_controller and not len( self.settings.mesh_children ):
            op( c, MAGICTRAILS_OT_new_trail_on_controller )
        self.draw_layout_convert( c )

    def draw_layout_convert( self, layout ):

        overrides = {'object' : self.id_data }
        if self.type == 'NONE':
            op( layout, MAGICTRAILS_OT_convert, text = 'Make Controller', icon = 'OUTLINER_OB_EMPTY', overrides = overrides ).convert_type = 'CONTROLLER'
        else:
            op( layout, MAGICTRAILS_OT_convert, text = 'Disconnect', icon = 'UNLINKED', overrides = overrides ).convert_type = 'NONE'

def register_properties( register ):

    if register:
        Object.mtrail = PointerProperty( type = MagicTrailsProperty )
        Object.speed = FloatProperty( )
        Scene.mtrail_is_baking = BoolProperty( default = False )
        Scene.mtrail_settings = PointerProperty( type = TrailDefaults )

    else:
        del Object.mtrail
        del Object.speed
        del Scene.mtrail_is_baking
        del Scene.mtrail_settings

def get_classes( ):

    yield TrailDefaults
    yield MagicTrailsProperty