import bpy
from bpy.types import PropertyGroup, Object
from bpy.props import *
from mathutils import Vector, Matrix
import numpy as np
from time import perf_counter as pc

from .settingutils import set_controller, update_is_baked ,material_group_label
from .trailsettings import MagicTrailsSettings

from ..trail_generator import setup_trail_armature, arm_mod_name, subsurf_name
from ..operators import *
from ..operators import draw_operator as op

def t( ):
    return pc( ) * 10 ** 6

def d( start, stop, m = 1 ):
    dur = t( ) - start
    print( 'time:', dur * m )

class TrailMeshSettings( MagicTrailsSettings, PropertyGroup ):

    controller : PointerProperty( type = Object )
    is_baked : BoolProperty( name = 'Is Baked', default = False, update = update_is_baked )
    bake_armature : PointerProperty( type = Object )

    @property
    def controller( self ):
        return self.controller
    
    @controller.setter
    def controller( self, value ):
        set_controller( self, value )

    @property
    def length( self ):
        return int( len( self.id_data.data.vertices ) / 2 )

    def set_mesh_segment( self, segment, matrix ):
        
        mesh = self.id_data.data
        v1 = mesh.vertices[ segment * 2 ]
        v2 = mesh.vertices[ segment * 2 + 1 ]
        v1.co = matrix.translation
        loc2 = matrix @ Vector(( 0, 0, 1 ))
        v2.co = loc2

    def reset_position( self ):

        for i in range( self.length ):
            self.set_mesh_segment( i, Matrix( ))

    def realtime_update_old( self ):
        
        if self.is_baked:
            return
        mesh_len = self.length
        for i in range( mesh_len ):
            mat = self.controller.mtrail.settings.get_data( i )
            self.set_mesh_segment( i, mat )
    
    def realtime_update( self ):
         
        if self.is_baked:
            return
        if not self.controller:
            print( f'No Controller found for "{self.id_data.name}". Can not move.' )
        length = self.length
        a = np.zeros( length * 6 )
        cache = self.controller.mtrail.settings.get_cache( )
        for i in range( length ):
            mat = self.get_data( cache, i )
            root = mat.translation
            tip = mat @ Vector(( 0, 0, 1 ))
            e = enumerate( list( root ) + list( tip ))
            for k, v in e:
                a[ 6*i + k ] = v
        
        self.id_data.data.vertices.foreach_set( 'co', a )
        self.id_data.data.update( )

    def bake_in_range( self, start, end ):
        
        obj = self.id_data
        arm_obj = setup_trail_armature( bpy.context, f'{self.controller.name}_armature', self.id_data )
        bones = arm_obj.pose.bones
        self.is_baked = True
        scene = bpy.context.scene
        frame_mem = scene.frame_current
        for f in range( start, end + 1 ):
            scene.frame_set( f )
            for offset, b in enumerate( bones ):
                mat = self.controller.mtrail.settings.get_data( offset )
                b.location = mat.translation
                b.rotation_quaternion = mat.to_quaternion( )
                b.scale = mat.to_scale( )
                for t in ( 'location', 'rotation_quaternion', 'scale' ):
                    b.keyframe_insert( t, frame = scene.frame_current )

        print( f'baked {end +1 -start } frames for object "{obj.name}"' )
        scene.frame_set( frame_mem )

    def clear_bake( self ):
        
        self.is_baked = False
        obj = self.id_data
        mods = obj.modifiers
        mod = mods[ arm_mod_name ]
        arm_obj = mod.object
        mods.remove( mod )
        bpy.data.armatures.remove( arm_obj.data )
    
    def draw_layout( self, layout, show_options = False, draw_children = False, box = True ):

        obj = self.id_data
        if box:
            layout = layout.box( )
        c = layout.column( align = False )
        self.draw_info( c, icon_value = icon( 'TRAIL_MESH' ))
        #material change button
        op( c, MAGICTRAILS_OT_set_material, text = obj.data.materials[0].name, icon = "MATERIAL", overrides = { 'object' : obj })

        subsurf = obj.modifiers.get( subsurf_name )
        if show_options:
            if subsurf:
                split = c.split( align = True )
                split.label( text = 'Smooth:')
                split.prop( subsurf, 'levels', text = '' )
                split.prop( subsurf, 'render_levels', text = '' )
            self.draw_bake( c, cls = MAGICTRAILS_OT_bake_to_armature )
            self.draw_material_properties( c )

    def update_material( self, mat ):

        obj = self.id_data
        obj.data.materials.clear( )
        new_mat_name = f'{mat.name} [{obj.name}]'
        if new_mat_name in bpy.data.materials.keys( ):
            bpy.data.materials.remove( bpy.data.materials[ new_mat_name ])
        mat.use_fake_user = True
        mat_instance = mat.copy( )
        mat_instance.name = new_mat_name
        obj.data.materials.append( mat_instance )

    def draw_material_properties( self, layout ):

        obj = self.id_data
        inputs = []
        for m in ( x for x in obj.data.materials if x.node_tree ):
            for n in ( x for x in m.node_tree.nodes if x.bl_idname == 'ShaderNodeGroup' and x.label == material_group_label ):
                for i in ( x for x in n.inputs if x.name.startswith( ':' )):
                    inputs.append( i )
        if len( inputs ):        
            box = layout.box( )
            c = box.column( align = True, heading = 'Material Properties' )
            for i in inputs:
                c.prop( i, 'default_value', text = i.name )

    def free( self, delete_children = False ):
        
        if self.is_baked:
            obj = self.id_data.modifiers[ arm_mod_name ].object
            if obj:
                bpy.data.armatures.remove( obj.data )
        bpy.data.meshes.remove( self.id_data.data )

def get_classes( ):
    yield TrailMeshSettings