import bpy, numpy as np
from math import floor, pi
from .utils import get_trail_collection

subsurf_name = 'magic_trails_subsurf'
arm_mod_name = 'magic_trails_armature'

def generate_trail_mesh( res, name = 'magic_trail_mesh' ):

    cuts = res - 1
    verts = [ ]
    faces = [ ]
    uvs = [ ]

    for i in ( 0, 1 ):
        verts.append(( 0, 0, i ))
        uvs.append(( 0, i ))
    for i in range( cuts ):
        frac = ( i + 1 ) / cuts
        for k in ( 0, 1 ):
            verts.append(( 0, 0, k ))
            uvs.append(( frac, k ))
        a = np.array(( 0,2,3,1 ))
        faces.append( a + 2*i )
    
    mesh = bpy.data.meshes.new( name )
    mesh.from_pydata( verts, [], faces )

    uv_data = mesh.uv_layers.new( ).data
    for face in mesh.polygons:
        face.use_smooth = True
        for vert_idx, loop_idx in zip(face.vertices, face.loop_indices):
            uv_data[loop_idx].uv = ( floor(vert_idx/2)/cuts, vert_idx % 2 )
    bpy.context.view_layer.objects.active = bpy.data.objects.get(name)
    bpy.ops.object.shade_smooth()
    return mesh

def setup_trail( c_object, settings ):

    controller = c_object
    obj = None

    controller.mtrail.convert_to( 'CONTROLLER' )
    controller.empty_display_size = 1
    controller.empty_display_type = 'SINGLE_ARROW'
    c_settings = controller.mtrail.settings
    c_settings.cache_length = settings.trail_resolution

    if not settings.generate_mesh:
        print( 'New Trail without Trail Mesh' )
    else:
        mesh = generate_trail_mesh( settings.trail_resolution, f'{settings.name}_mesh' )

        obj = bpy.data.objects.new( f'{settings.name}_mesh', mesh )

        subsurf = obj.modifiers.new( subsurf_name, 'SUBSURF' )
        subsurf.boundary_smooth = 'PRESERVE_CORNERS'
        subsurf.quality = 1
        subsurf.levels = 1
        subsurf.render_levels = 2
        subsurf.show_viewport = True

        obj.mtrail.convert_to( 'MESH', controller = controller )
        obj.mtrail.settings.update_material( settings.get_material( ))
    
    coll = get_trail_collection( )

    for o in ( controller, obj ):
        if o and not o.name in bpy.context.scene.objects.keys( ):
            coll.objects.link( o )

def generate_trail_armature( context, name, mesh ):

    mem_active = context.object
    mem_mode = mem_active.mode
    bpy.ops.object.mode_set( mode = 'OBJECT', toggle = False )

    arm = bpy.data.armatures.new( name )
    arm_obj = bpy.data.objects.new( name, arm )
    get_trail_collection( ).objects.link( arm_obj )

    bpy.context.view_layer.objects.active = arm_obj
    bpy.ops.object.mode_set( mode = 'EDIT', toggle = False )

    eb = arm.edit_bones
    trail_len = int( len( mesh.vertices ) / 2 )
    for i in range( trail_len ):
        b = eb.new( f'seg_{i}' )
        b.head = ( 0, 0, 0 )
        b.tail = ( 0, 1, 0 )
    bpy.ops.object.mode_set( mode = 'OBJECT', toggle = False )
    
    bpy.context.view_layer.objects.active = mem_active
    bpy.ops.object.mode_set( mode = mem_mode, toggle = False )

    arm_obj.hide_render = True
    arm_obj.hide_viewport = True
    return arm_obj

def setup_trail_armature( context, name, mesh_obj ):

    groups = mesh_obj.vertex_groups
    arm_obj = generate_trail_armature( context, name, mesh_obj.data )
    arm = arm_obj.data
    pbones = arm.bones
    for i, pb in enumerate( list( pbones )):
        name = pb.name
        if name in groups:
            groups.remove( groups[ name ])
        g = groups.new( name = name )
        g.add( [ 2*i, 2*i + 1 ], 1, 'REPLACE' )
    
    arm_mod = mesh_obj.modifiers.new( arm_mod_name, 'ARMATURE' )
    arm_mod.object = arm_obj
    arm_mod.use_vertex_groups = True
    bpy.ops.object.modifier_move_to_index( {'object': mesh_obj }, modifier = arm_mod.name, index = 0 )
    return arm_obj